#include "gl.h"
#include "fastmath.h"
#include "world.h"
#include "glbox.h"

#include "terrain.h"

static GLFix xr = 0, yr = 0;
static const GLFix incr = 20;

static void getForward(GLFix *x, GLFix *z)
{
    *x = fast_sin(yr) * incr;
    *z = fast_cos(yr) * incr;
}

static void getRight(GLFix *x, GLFix *z)
{
    *x = fast_sin((yr + 90).normalise_angle()) * incr;
    *z = fast_cos((yr + 90).normalise_angle()) * incr;
}

static void drawTexture(int screen_x, int screen_y, int w, int h, TEXTURE *texture, int tex_x, int tex_y, COLOR *buffer)
{
    for(int y = 0; y < h; y++)
        for(int x = 0; x < w; x++)
            buffer[screen_x + x + (screen_y + y)*WIDTH] = texture->bitmap[tex_x + x + (tex_y + y)*texture->width];
}

#define CR_PIXEL(x, y) (screen_buffer[WIDTH/2+(x) + (HEIGHT/2+(y))*WIDTH] = ~screen_buffer[WIDTH/2+(x) + (HEIGHT/2+(y))*WIDTH])

int main()
{
    if(!lcd_isincolor())
    {
        show_msgbox("Error", "Your LCD hasn't got enough colors :-P");
        return EXIT_FAILURE;
    }

    nglInit();

    COLOR* const screen_buffer = new COLOR[WIDTH*HEIGHT];
    nglSetBuffer(screen_buffer);

    init_blockData(&terrain);
    glBindTexture(&terrain);

    glLoadIdentity();

    int current_block = BLOCK_STONE;
    GLFix x = 0, y = BLOCK_SIZE * Chunk::SIZE + BLOCK_SIZE, z = 0;
    const GLFix player_width = BLOCK_SIZE*0.8f, player_height = BLOCK_SIZE*1.8f, eye_pos = BLOCK_SIZE*1.6f;
    bool finished = false, key_held_down = false, can_jump = false, tp_had_contact = false;
    int tp_last_x = 0, tp_last_y = 0;
    GLFix vy = 0; //Y-Velocity for gravity and jumps

    World world;

    while(!finished)
    {
        //printf("X: %d Y: %d Z: %d XR: %d YR: %d\n", x.to_int(), y.to_int(), z.to_int(), xr.to_int(), yr.to_int());

        glColor3f(0.1f, 0.1f, 1); //Blue background
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glPushMatrix();

        nglRotateX((GLFix(360) - xr.normalise_angle()).normalise_angle());
        nglRotateY((GLFix(360) - yr.normalise_angle()).normalise_angle());
        glTranslatef(-x, -y - eye_pos, -z);

        world.setPosition(x, y, z);

        world.render();

        glPopMatrix();

        //Draw GUI

        CR_PIXEL(0, 0);
        CR_PIXEL(-1, 0);
        CR_PIXEL(-2, 0);
        CR_PIXEL(0, -1);
        CR_PIXEL(0, -2);
        CR_PIXEL(1, 0);
        CR_PIXEL(2, 0);
        CR_PIXEL(0, 1);
        CR_PIXEL(0, 2);

        TextureAtlasEntry &block = block_textures[current_block];
        drawTexture(0, 0, 16, 16, &terrain, block.left, block.top, screen_buffer);

        //Finished drawing GUI

        nglDisplay();

        if(isKeyPressed(KEY_NSPIRE_ESC))
            break;

        if(isKeyPressed(KEY_NSPIRE_7)) //Put block down
        {
            Position res; AABB::SIDE side;
            GLFix dx = fast_sin(yr)*fast_cos(xr), dy = -fast_sin(xr), dz = fast_cos(yr)*fast_cos(xr);
            if(world.intersectRay(x, y + eye_pos, z, dx, dy, dz, res, side))
            {
                switch(side)
                {
                case AABB::BACK:
                    ++res.z;
                    break;
                case AABB::FRONT:
                    --res.z;
                    break;
                case AABB::LEFT:
                    --res.x;
                    break;
                case AABB::RIGHT:
                    ++res.x;
                    break;
                case AABB::BOTTOM:
                    --res.y;
                    break;
                case AABB::TOP:
                    ++res.y;
                    break;
                default:
                    break;
                }
                world.setBlock(res.x, res.y, res.z, current_block);
            }
        }
        else if(isKeyPressed(KEY_NSPIRE_9)) //Remove block
        {
            Position res; AABB::SIDE side;
            GLFix dx = fast_sin(yr)*fast_cos(xr), dy = -fast_sin(xr), dz = fast_cos(yr)*fast_cos(xr);
            if(world.intersectRay(x, y + eye_pos, z, dx, dy, dz, res, side))
                world.setBlock(res.x, res.y, res.z, BLOCK_AIR);
        }

        GLFix dx = 0, dz = 0;

        if(isKeyPressed(KEY_NSPIRE_8)) //Forward
        {
            GLFix dx1, dz1;
            getForward(&dx1, &dz1);

            dx += dx1;
            dz += dz1;
        }
        else if(isKeyPressed(KEY_NSPIRE_2)) //Backward
        {
            GLFix dx1, dz1;
            getForward(&dx1, &dz1);

            dx -= dx1;
            dz -= dz1;
        }

        if(isKeyPressed(KEY_NSPIRE_4)) //Left
        {
            GLFix dx1, dz1;
            getRight(&dx1, &dz1);

            dx -= dx1;
            dz -= dz1;
        }
        else if(isKeyPressed(KEY_NSPIRE_6)) //Right
        {
            GLFix dx1, dz1;
            getRight(&dx1, &dz1);

            dx += dx1;
            dz += dz1;
        }

        AABB aabb(x - player_width/2, y, z - player_width/2, x + player_width/2, y + player_height, z + player_width/2);

        if(!world.intersect(aabb))
        {
            AABB aabb_moved = aabb;
            aabb_moved.low_x += dx;
            aabb_moved.high_x += dx;
            aabb_moved.low_z += dz;
            aabb_moved.high_z += dz;

            if(!world.intersect(aabb_moved))
            {
                x += dx;
                z += dz;

                aabb = aabb_moved;
            }

            aabb.low_y += vy;
            aabb.high_y += vy;

            can_jump = world.intersect(aabb);

            if(!can_jump)
                y += vy;
            else
                vy = 0;

            vy -= 5;
        }

        if(isKeyPressed(KEY_NSPIRE_5) && can_jump) //Jump
        {
            vy = 50;
            can_jump = false;
        }

        touchpad_report_t touchpad;
        touchpad_scan(&touchpad);

        if(touchpad.pressed)
        {
            switch(touchpad.arrow)
            {
            case TPAD_ARROW_DOWN:
                xr += incr/3;
                break;
            case TPAD_ARROW_UP:
                xr -= incr/3;
                break;
            case TPAD_ARROW_LEFT:
                yr -= incr/3;
                break;
            case TPAD_ARROW_RIGHT:
                yr += incr/3;
                break;
            case TPAD_ARROW_RIGHTDOWN:
                xr += incr/3;
                yr += incr/3;
                break;
            case TPAD_ARROW_UPRIGHT:
                xr -= incr/3;
                yr += incr/3;
                break;
            case TPAD_ARROW_DOWNLEFT:
                xr += incr/3;
                yr -= incr/3;
                break;
            case TPAD_ARROW_LEFTUP:
                xr -= incr/3;
                yr -= incr/3;
                break;
            }
        }
        else if(tp_had_contact && touchpad.contact)
        {
            yr += (touchpad.x - tp_last_x) / 50;
            xr -= (touchpad.y - tp_last_y) / 50;
        }

        tp_had_contact = touchpad.contact;
        tp_last_x = touchpad.x;
        tp_last_y = touchpad.y;

        if(isKeyPressed(KEY_NSPIRE_1) && !key_held_down) //Switch block type
        {
            --current_block;
            if(current_block < 0)
                current_block = BLOCK_MAX - 1;

            key_held_down = true;
        }
        else if(isKeyPressed(KEY_NSPIRE_3) && !key_held_down)
        {
            ++current_block;
            if(current_block == BLOCK_MAX)
                current_block = 1;

            key_held_down = true;
        }
        else if(key_held_down)
            key_held_down = false;
    }

    nglUninit();

    delete[] screen_buffer;

    return EXIT_SUCCESS;
}
